package se.umu.jobsubmission.impl.broker; 

import java.util.Calendar;
import java.util.Map;

import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.globus.mds.glue.GLUECERPType;
import org.globus.mds.glue.StateType;
import org.globus.wsrf.client.ServiceURL;

import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequirementType;
import se.umu.jobsubmission.util.JSDLUtils;
import se.umu.jobsubmission.util.ReservationUtils;
import se.umu.reservation.ReservationType;
import se.umu.wsag.client.ReservationClient;
import se.umu.wsag.impl.AgreementFactoryConstants;
import se.umu.wsag.util.JobSubmissionUtils;

import org.ggf.ws_agreement.ContinuingFaultType;

import org.globus.mds.glue.ComputingElementType;

public class StartTimePredictor implements Predictor {
	
	
	private static Logger logger = Logger.getLogger(StartTimePredictor.class);
	
	// TODO: Make these configurable. Where? 
	private static final int PING_TIMEOUT = 2000;  // 2 seconds
	private static final int DEFAULT_LATESTSTART = 240; // 10 days
	private static final int DEFAULT_STARTTIME = 2880; // 2 days;
	
	private static final int FREE_CPU_FACTOR = 5;  	// Adjust queue waiting time
												    // by no more than 5 in order
													// to simulate backfilling.  
	private static final int AVERAGE_JOB_TIME = 300; // 5 hours
	private static final int AVERAGE_JOB_CPUS = 10; 
	
	private ReservationUtils reservationUtils; 
	
	private Predictor stageInTimePredictor; 
	private Predictor runTimePredictor; 
	
	public StartTimePredictor() {
		this.stageInTimePredictor = new StageInTimePredictor(); 
		this.runTimePredictor = new RunTimePredictor(); 
	}
	
	/**
	 * 
	 * @param resource GLUE object describing resource, including current load. 
	 * @return The predicted time the job has to wait in the queue. 
	 * @throws Exception
	 */
    public int predictFromLoad(GLUECERPType resource, int numberOfCPUs, 
    		int duration) throws Exception {    	
    	
    	logger.debug("Predicting queue time from load..."); 
    	try {
    		
    		ComputingElementType ce = resource.getComputingElement(0); 
    	
    		int totalJobs = ce.getState().getTotalJobs();      		
    		int totalCPUs = Integer.parseInt(ce.getInfo().getTotalCPUs()); 
    		int freeCPUs = ce.getState().getFreeCPUs(); 
    		
    		// Convert duration to minutes
    		duration = duration / 60; 
    		
    		logger.debug("TotalJobs: " + totalJobs); 
    		logger.debug("TotalCPUs: " + totalCPUs); 
    		logger.debug("freeCPUs: " + freeCPUs); 
    		logger.debug("Duration: " + duration); 
    		
    		// Well, this is not nice at all    		

    		// Assume that each job uses  CPU and 
    		// that we have FIFO scheduling...
    		int startTime =  (int) (StartTimePredictor.AVERAGE_JOB_TIME
    				* StartTimePredictor.AVERAGE_JOB_CPUS
    				* totalJobs / totalCPUs); 
    		logger.debug("start time (predicted from queue length): " 
    				+ startTime); 

    		// Now, compensate if there are free cpus, by adjusting the 
    		// start time by the % of the CPUs that are free. 
    		// The FREE_CPU_FACTOR dampens the effect of this. 
    		if(freeCPUs > 0) {    			
    			float f = ((float)totalCPUs) / 
					(totalCPUs + (StartTimePredictor.FREE_CPU_FACTOR*freeCPUs)); 
    			startTime = (int) (0.5 + startTime * f); 
    			logger.debug("start time (after free cpu adjustment): " + startTime); 
    		}
    		
    		// Do the same for start time, compensate for the 
    		// shorter (or longer) duration of our job compared to an average job.
    		startTime = (int) (startTime * Math.sqrt((double)duration / 
    				StartTimePredictor.AVERAGE_JOB_TIME)); 
    		logger.debug("start time (after run time adjustment): " + startTime); 
    		
    		// Compensate if this job uses fewer CPUs than the average job
    		double CPUfactor = Math.sqrt(((double) numberOfCPUs) 
    				/ StartTimePredictor.AVERAGE_JOB_CPUS); 
    		startTime = (int) (0.5 + startTime * CPUfactor);     		    		
    		logger.debug("start time (after #CPUs adjustment): " + startTime); 
    		
    		return startTime; 
    		
    	} catch (Exception e) {
    		logger.debug("Prediction failed: " + e.getMessage()); 
    		throw e; 
    	}
	}
    
	/**
	 * Creates a reservation and predicts (determines) start time 
	 * from start time of reservation. 
	 * @param resource
	 * @param earliestStart
	 * @param latestStart
	 * @param numberOfCPUs
	 * @param duration
	 * @param flexible
	 * @return The predicted start time. 
	 */
	public int predictUsingReservation(GLUECERPType resource, 
			Calendar earliestStart, 
			Calendar latestStart, 
			int numberOfCPUs, 
			int duration, 
			boolean flexible) throws Exception {
		
		String serviceHost = resource.getCluster(0).getSubCluster(0).getHost(0).getName(); 
			
    	ReservationClient reservationClient = new ReservationClient(); 
    	// Get host name of resource, and contacts its AgreementFactory
    	String agreementFactoryURI;
    	agreementFactoryURI = ServiceURL.getURL(serviceHost, 
    			AgreementFactoryConstants.serviceName).toString(); 
    	
    	// Use a ping call before trying to invoke service, 
    	// it takes forever for this client to determine that 
    	// no connection can be established to the service engine
    	if(!(JobSubmissionUtils.ping(serviceHost))) {
    		// logger.debug("Can not find service " + agreementFactoryURI); 
    		logger.debug("Resource " + serviceHost + 
    				" does not support reservations"); 
    		logger.debug("Using fallback method (load-based prediction)"); 
    		throw new Exception("Resource " + serviceHost + 
    				" does not support reservations"); 
    	}
    	
    	logger.debug("Creating reservation using\n" +agreementFactoryURI); 
    	EndpointReferenceType epr = reservationClient.createReservation(
                agreementFactoryURI, 
                numberOfCPUs, 
                earliestStart, 
                latestStart, 
                duration, 
                flexible);     	    	
    	
    	if(epr == null) {
    		logger.warn("Failed to create reservation..."); 
    		logger.debug("Using fallback method (load-based prediction)"); 
    		throw new Exception("Failed to create reservation");
    	}
    	
    	// Get exact start time of reservation...
    	ReservationType reservationInfo = reservationClient.
				getReservationProperties(epr); 
    	
    	// Store reservation for later use...
    	logger.debug("Storing reservation for later use..."); 
    	reservationUtils.storeReservation(epr, resource, reservationInfo); 
    	
    	Calendar exactStart = reservationInfo.getExactStart();
    	Calendar now = Calendar.getInstance(); 
    	long startDelay = 
    		(exactStart.getTimeInMillis() - now.getTimeInMillis()) / 1000;
    	
    	// Convert to minutes
    	startDelay = (59 + startDelay) / 60;     	
    	return (int) startDelay;     		
		
	}

	
	/**
	 * Method of attack. First (unless user requested otherwise), 
	 * create reservation and let job start time be reservation start time. 
	 * Should this fail, predict start time from current load. 
	 */
    public int predict(JobDefinition_Type jobDescription, 
    		Object resourceObject, 
			JobPreferencesType preferences, 
			Object context) {
        
    	GLUECERPType resource = (GLUECERPType) resourceObject; 
    	    	    
    	// Earliest and latest possible job start. 
    	Calendar earliestStart = Calendar.getInstance(); 
    	
    	// Time for input file staging, either specified by 
    	// user or predicted from file sizes. 
    	Map times = (Map) context; 
    	Integer tmpInteger; 
    	tmpInteger = (Integer) times.get("stageInTime"); 
    	int stageInDelay = tmpInteger.intValue(); 
    	earliestStart.add(Calendar.MINUTE, stageInDelay); 
    	
    	// ii) Latest start (may be specified by user. 
    	// Note that latestRelativeStart must be converted into
    	// latestAbsoluteStart before this call.         	
    	JobRequirementType reqs = null; 
    	Calendar latestStart; 
    	Calendar tmp = null; 
    	if(preferences != null)
    		reqs = preferences.getJobRequirement(); 
    	if(reqs != null)    		
    		tmp = reqs.getLatestAbsoluteStart();     	
    	if(tmp != null)
    		latestStart = tmp; 
    	else {
    		latestStart = Calendar.getInstance(); 
    		latestStart.add(Calendar.HOUR, StartTimePredictor.DEFAULT_LATESTSTART); 
    	}
    	logger.debug("LatestStart is: " + latestStart.getTime()); 
    	
    	// Duration of job (in seconds) 
    	tmpInteger = (Integer) times.get("runTime"); 
    	int duration = tmpInteger.intValue(); 
    	
    	logger.debug("From predict-context: stageIn: " + stageInDelay 
    			+ " duration: " + duration); 
    	
    	// Number of CPUs
    	int numberOfCPUs = JSDLUtils.getNumberOfCPUs(jobDescription); 
    	
    	// Job is flexible?
    	boolean flexible; 
    	try {
    		 flexible = preferences.getJobRequirement().
					getFlexible().booleanValue(); 
    	} catch (Exception e) {
    		flexible = false; 
    	}
 
    	// Should a reservation be used for this job. 
    	boolean useReservation; 
    	try {
    		useReservation = preferences.getJobRequirement().
					getUseReservations().booleanValue(); 
    	} catch (Exception e) {
    		useReservation = true; 
    	}
    	
    	// Create a reservation and let reservation start 
    	// time be job start time. 
    	if(useReservation) {
    		try {
    			int startTime =  predictUsingReservation(resource, 
    					earliestStart, 
						latestStart, 
						numberOfCPUs, 
						duration + 120, // Add 2 extra minutes, just in case...
						flexible); 
    			return startTime; 
    		} catch (ContinuingFaultType f) {
    			logger.info("Reservation failed but is possible later..."); 
    		} catch (Exception e) {    		
    			logger.debug("Reservation not possible, " + 
    			"predicting start time from load"); 
    		}
    	}    	
    	
    	// Note that we have to add stageInDelay in the last two 
    	// cases for consistency. 
    	// (it is added when we predict start time from reservations) 
    	
    	// Use fallback methods, that is, predict from load
    	try {
    		int startTime = predictFromLoad(resource, 
    				numberOfCPUs, 
					duration); 
    		return startTime + stageInDelay; 
    	} catch (Exception e) {
    		logger.debug("Could not determine start time from load " 
    				+ " using default value"); 
    	}
    	
    	// Use default time
    	logger.debug("Using default start time (+delay)" 
    			+ StartTimePredictor.DEFAULT_STARTTIME + 
    					"(" + stageInDelay + ")"); 
    	
    	return StartTimePredictor.DEFAULT_STARTTIME + stageInDelay; 
    	
    }
    
    public void setReservationUtils(ReservationUtils utils) {
    	this.reservationUtils = utils; 
    }

}
