package se.umu.jobsubmission.impl.jobsubmissionservice; 

import java.rmi.RemoteException;
import java.io.StringWriter; 

import javax.xml.rpc.ServiceException;
import javax.xml.rpc.server.ServiceLifecycle;

import org.apache.axis.MessageContext;
import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.wsrf.utils.ContextUtils;
import org.oasis.wsrf.faults.BaseFaultType;

import org.globus.wsrf.encoding.ObjectSerializer;

import se.umu.jobsubmission.InvalidJobDescriptionFaultType;
import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequest;
import se.umu.jobsubmission.JobRequestFaultType;
import se.umu.jobsubmission.NoAptResourcesFoundFaultType;
import se.umu.jobsubmission.NoResourcesFoundFaultType;
import se.umu.jobsubmission.SubmittedJob;
import se.umu.jobsubmission.impl.submitter.Submitter;
import se.umu.jobsubmission.impl.submitter.Coallocator;

import se.umu.jobsubmission.CoallocateJobResponse;
import se.umu.jobsubmission.CoallocationRequest; 

import org.globus.wsrf.utils.FaultHelper;


import org.globus.wsrf.ResourceKey;
import org.ggf.ws_agreement.ContinuingFaultType;
import org.globus.wsrf.utils.AddressingUtils;
import org.apache.axis.message.addressing.EndpointReferenceType;


/**
 * Service for brokering (scheduling) of Grid jobs. 
 * Service recieves job requests containing JSDL job description, 
 * list of custom index servers to query for resources, 
 * and optionally, a document describing job preferences. 
 */

public class JobSubmissionService implements ServiceLifecycle {

    private static Logger logger = Logger.getLogger(JobSubmissionService.class); 
    private Submitter submitter;
    private Coallocator coallocator; 
    

    public JobSubmissionService() {
    }

    // Specified by ServiceLifecycle
    public void destroy() {
        logger.debug("Destroy invoked"); 
        submitter.shutdown(); 
    }
    
    // Invoked upon creation of service, initiates 
    // xml driver and configures submitter. 
    public void init(java.lang.Object context) throws ServiceException {

        try {

            MessageContext ctx = MessageContext.getCurrentContext(); 

            // Configure submitter
            String configFile = (String) ContextUtils.getServiceProperty(ctx, 
                    JobSubmissionConstants.SUBMITTER_CONFIG_FILE); 
            String globusLocation = ContainerConfig.getGlobusLocation(); 
            String filePath = globusLocation + configFile; 

            logger.debug("JSS config file: " + filePath); 
            submitter = new Submitter(filePath); 
            
            coallocator = new Coallocator(submitter.getInformationFinderFactory(),
            		submitter.getDispatcherClasses(),
					submitter.getDataManagerClasses(),
					submitter.getNumberOfThreads()); 
                    
            
        } catch (Exception e) {
            throw new ServiceException("Error configuring service", e); 
        }
        
    }
    
  
    /**
     * Method specified by wsdl porttype
     * @param param Contains the job description and optionally, 
     * job preferences and index servers to query for resources (instead
     * of using the default ones). 
     * @return Job ID as a string (can be EPR or other kind of URI)
     * @throws RemoteException
     * @throws JobRequestFaultType
     */
    public String requestJob(JobRequest request) 
    	throws RemoteException, 
		BaseFaultType, 
		NoResourcesFoundFaultType, 
		InvalidJobDescriptionFaultType, 
		JobRequestFaultType, 
		NoAptResourcesFoundFaultType { 

        logger.info("New job requested"); 
    	
        SubmittedJob jobInfo = null; 
        
        JobDefinition_Type jobDesc = null; 
        
        // Submit job. 
        try {
        	
            // JobRequest request = param.getRequest(); 
            if(request == null) 
                throw new RemoteException("No job request specified!"); 
            
            String host = request.getClientURL(); 
            logger.info("Job submitted from host: " + host); 

            jobDesc = request.getJobDescription(); 
  
            Boolean clientStages = request.getClientStagesFiles(); 
            boolean clientStagesInput = true; 
            if(clientStages != null && clientStages.booleanValue() == false) 
            	clientStagesInput = false; 
            
            String indexServerURIs[] = request.getIndexServerURIs(); 
            JobPreferencesType jobPref = request.getJobPreferences(); 
            String clientURL = request.getClientURL(); 
            
            StringWriter sw = new StringWriter(); 
            ObjectSerializer.serialize(sw, jobDesc, JobDefinition_Type.getTypeDesc().getXmlType()); 
            logger.debug("Job description is: " + sw.toString());             
            logger.debug("Client URL: " + clientURL);  
            logger.debug("Index servers: " ); 
            if(indexServerURIs != null) {
            	for(int i = 0; i < indexServerURIs.length; i++) {
            		logger.debug(indexServerURIs[i]); 
            	}             
            } 
            
            jobInfo = submitter.submitJob(jobDesc, 
            		indexServerURIs, 
					jobPref, 
					clientURL, 
					clientStagesInput);                         
          
        } 
        catch (RemoteException re) {
        	logger.error("Job request failed (RemoteException): " 
        			+ re.getMessage());  
        	throw re; 
        }
        catch (Exception e) {
        	logger.error("Job request failed: " + e.getMessage()); 
        	throw new RemoteException(e.getMessage()); 
        }
     
        // Get resource home and create new resource. 
    	ResourceContext ctx = null;
		JobSubmissionHome home = null;
		
		// Create resource and add job info as resource property, 
		// synchronize this as more than one job may be submitted simultaneously
		// FIXME: Enought to synchronize the home.create(jobInfo) call? 
		// Maybe, but this is not a bottleneck as far as I understand...
		synchronized (this) {
			try {			
				ctx = ResourceContext.getResourceContext();
				home = (JobSubmissionHome) ctx.getResourceHome();			
				home.create(new SubmittedJob[]{jobInfo}); 						
			} catch (Exception e) {
				logger.error("Could not create resource"); 
				throw new RemoteException("Could not create resource", e); 
			}
		}

		String localID = jobInfo.getLocalID(); 
		logger.info("Job submission done, returning job id: " + localID); 
        return localID; 
    }
    
    /**
     * CoAllocation (actually, coreserves), a number of job descriptions
     * within a given allowed start time window and with a maximum
     * allowed subjob start time deviation. 
     * @param parameters
     * @return
     * @throws RemoteException
     * @throws NoResourcesFoundFaultType
     * @throws BaseFaultType
     * @throws InvalidJobDescriptionFaultType
     * @throws JobRequestFaultType
     * @throws NoAptResourcesFoundFaultType
     */
    public CoallocateJobResponse coallocateJob(CoallocationRequest request) 
    	throws RemoteException, 
		NoResourcesFoundFaultType, 
		BaseFaultType, 
		InvalidJobDescriptionFaultType, 
		JobRequestFaultType, 
		NoAptResourcesFoundFaultType {
    	
    	logger.info("New coallocation requested");       	
    	
        JobDefinition_Type jobDescriptions[] = null; 
        SubmittedJob jobs[] = null; 
        // Submit job. 
        try {
        	
            // CoallocationRequest request = arg.getRequest(); 
            if(request == null) 
                throw new RemoteException("No job request specified!"); 
            
            String clientHost = request.getClientURL(); 
            logger.info("Job submitted from host: " + clientHost); 
            
            int tolerance = Coallocator.DEFAULT_TOLERANCE; 
            Integer userTolerance = request.getTolerance(); 
            if(userTolerance != null && userTolerance.intValue() > 0) {
            	tolerance = userTolerance.intValue(); 
            }

            jobDescriptions = request.getJobDescription(); 
            
            Boolean clientStages = request.getClientStagesFiles(); 
            boolean clientStagesInput = true; 
            if(clientStages != null && clientStages.booleanValue() == false) 
            	clientStagesInput = false; 
            
            String indexServerURIs[] = request.getIndexServerURIs(); 
            JobPreferencesType jobPreferences[] = request.getJobPreferences(); 
            String clientURL = request.getClientURL(); 
            
            jobs = coallocator.coallocateJob(jobDescriptions, 
                    indexServerURIs, 
                    jobPreferences, 
                    clientHost, 
        			clientStagesInput, 
					tolerance);    		
    		
    	} catch (ContinuingFaultType f) {
    		logger.error("Coallocation attempt failed as all subjobs" +
    				" could not be started within the given window"); 
    		throw f; 
    	} catch (BaseFaultType f) {
    		FaultHelper helper = new FaultHelper(f); 
    		logger.error("Coallocation failed: " + helper.getDescriptionAsString()); 
    		throw f;     		
    	}    	
    	catch (Exception e) {
    		logger.error("Coallocation attempt failed: " + e.getMessage());     	
    	 	JobRequestFaultType fault = new JobRequestFaultType(); 
        	FaultHelper helper = new FaultHelper(fault); 
        	helper.addDescription("Coallocation failed: " + e.getMessage()); 
        	throw helper.getFault(); 
    	}
    	
    	CoallocateJobResponse response = new CoallocateJobResponse(); 
    	String jobIds[] = new String[jobs.length]; 
    	for(int i = 0; i < jobs.length; i++) {
    		jobIds[i] = jobs[i].getLocalID(); 
    	}
    	response.setJobIds(jobIds); 
  
    	// Get resource home and create new resource. 
    	// Add resource epr to response. 
    	ResourceContext ctx = null;
		JobSubmissionHome home = null;
    	synchronized (this) {
			try {			
				ctx = ResourceContext.getResourceContext();
				home = (JobSubmissionHome) ctx.getResourceHome();			
				ResourceKey key = home.create(jobs); 	
				EndpointReferenceType epr = AddressingUtils.createEndpointReference(ctx, key); 
				response.setEndpointReference(epr); 
			} catch (Exception e) {
				logger.error("Could not create resource"); 
				throw new RemoteException("Could not create resource", e); 
			}
		}
    	
    	return response; 
    	
    }
    
}

