package se.umu.jobsubmission.impl.submitter; 

import java.io.FileInputStream;
import java.util.Calendar;
import java.util.List;
import java.util.Vector;

import javax.xml.namespace.QName;

import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl._2005._11.jsdl.JobDefinition_Type;
import org.globus.mds.glue.GLUECERPType;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.utils.FaultHelper;
import org.xml.sax.InputSource;

import se.umu.jobsubmission.JobPreferencesType;
import se.umu.jobsubmission.JobRequestFaultType;
import se.umu.jobsubmission.NoAptResourcesFoundFaultType;
import se.umu.jobsubmission.SubmittedJob;
import se.umu.jobsubmission.SubmitterConfiguration;
import se.umu.jobsubmission.impl.broker.Broker;
import se.umu.jobsubmission.impl.broker.Selector;
import se.umu.jobsubmission.impl.datamanager.DataManager;
import se.umu.jobsubmission.impl.dispatcher.Dispatcher;
import se.umu.jobsubmission.impl.informationfinder.InformationFinder;
import se.umu.jobsubmission.impl.informationfinder.InformationFinderFactory;
import se.umu.jobsubmission.impl.informationfinder.ResourceInformationCache;
import se.umu.jobsubmission.util.ReservationUtils;
import se.umu.jobsubmission.util.SubmitterUtils;
import se.umu.reservation.ReservationInfo;

/*
   The sumbmitter is responsible for coordinating the job submission process. 
   It uses a informationFinder, the Broker, the Dispatcher an the DataManager. 
   
   An outline of Submitter.submitJob(jobDesc): 

   Broker.verify(jobDesc);  

   glueInfo = resources = informationFinder.getResourceInfo(indexServers); 

   List possibleResources = Broker.filter(glueInfo, jobDesc); 
   List rankedResources = Broker.predict(possibleResources, jobDesc); 

   boolean done = false; 
   String jobContact; 
   Resource resource; 
   while(!done) {
       try {
           done = true; 
           resource = Broker.select(rankedResources, jobDesc);
           DataManager.stageIn(jobDesc, resource); 
           jobContact = Dispatcher.requestJob(resource, jobDesc, 
               resource.hasReservation()? resource.getReservation() : null); 
       } catch (Exception e) {
           Broker.discardResource(resource); 
           done = false; 
       }
   }

   return jobContact; 
 */

public class Submitter {


    private static Logger logger = Logger.getLogger(Submitter.class); 

    private InformationFinderFactory informationFinderFactory; 
    private String dispatcherClasses[]; 
    private String dataManagerClasses[]; 
    private int numberOfThreads; 
    
    private static final int DEFAULT_CONNECTION_TIMEOUT = 10000; 
    private static final int DEFAULT_THREAD_POOL_SIZE = 4; 

    private ResourceInformationCache resourceInformationCache; 

    /* Set up submitter from config file. */
    public Submitter(String configFile) throws Exception {
        logger.debug("Configuring submitter from file: " + configFile); 
        this.resourceInformationCache = new ResourceInformationCache(); 
        readConfig(configFile); 
    }

    /* Config file determines which dispatchers and infofinders to use. 
       Furthermore, the finders, fetchers and converters to use in the
       infofinder are configured as well as timeouts, threadpoolsize
       and default index server for the information finder is determined. 
       Each submission uses a separate informationfinder and dispatcher
       in order to avoid concurrency issues. */
    public void readConfig(String configFile) throws Exception {

        SubmitterConfiguration config = null;
        
        QName configQName = SubmitterConfiguration.getTypeDesc().getXmlType(); 
        FileInputStream input = new FileInputStream(configFile); 
        InputSource inputSource = new InputSource(input); 
        config = (SubmitterConfiguration) ObjectDeserializer.deserialize(inputSource, 
        		SubmitterConfiguration.class); 
        
		this.dispatcherClasses = config.getDispatcherClass(); 
		this.dataManagerClasses = config.getDataManagerClass(); 

        String finderClasses[] = config.getResourceFinderClass(); 
        String fetcherClasses[] = config.getInformationFetcherClass(); 
        String converterClasses[] = config.getInformationConverterClass(); 
        int timeout = config.getConnectionTimeout().intValue(); 
        if(timeout == 0) {
            logger.info("No timeout specified, using default (" + 
            		Submitter.DEFAULT_CONNECTION_TIMEOUT + ")"); 
            timeout = Submitter.DEFAULT_CONNECTION_TIMEOUT; 
        }
        int finderThreads = config.getThreadPoolSize().intValue(); 
        if(finderThreads == 0) {
            logger.info("Number of threads for specified, using default (" + 
            		Submitter.DEFAULT_THREAD_POOL_SIZE + ")"); 
            finderThreads = Submitter.DEFAULT_THREAD_POOL_SIZE; 
        }
        this.numberOfThreads = finderThreads; 

        String indexServers[] = config.getDefaultIndexServer(); 
        List indexServersList = new Vector(); 
        for(int i = 0; i < indexServers.length; i++) 
            indexServersList.add(indexServers[i]); 
        
        informationFinderFactory = new InformationFinderFactory(finderThreads, 
                finderClasses, 
                fetcherClasses, 
                converterClasses, 
                timeout, 
                indexServersList, 
                this.resourceInformationCache); 
    }
    
    /**
     * Ensure that job starts early enough if latestStart is specified
     * in the job preferences. 
     * @param reservation
     * @param reservationUtils
     * @throws Exception
     */
    public void assertLatestStart(ReservationInfo reservation, 
    		ReservationUtils reservationUtils, 
			Calendar latestStart) throws Exception {
    	logger.debug("Checking that start is not to late..."); 
		
		// No reservation -> no chance job will start in time
		if(reservation == null) {
			logger.debug("Job start is to late!"); 
			reservationUtils.removeUnusedReservations(); 
			NoAptResourcesFoundFaultType f = new NoAptResourcesFoundFaultType(); 
			FaultHelper helper = new FaultHelper(f); 
			helper.addDescription("Job start is too late"); 
			throw helper.getFault(); 
		}
		// If we have a reservation, ensure it starts 
		// early enough...
		Calendar resStart = reservation.getReservation().getEarliestStart(); 
		if(resStart.after(latestStart)) {
			logger.debug("Job start is too late!"); 
			reservationUtils.removeUnusedReservations(); 
			NoAptResourcesFoundFaultType f = new NoAptResourcesFoundFaultType(); 
			FaultHelper helper = new FaultHelper(f); 
			helper.addDescription("Earliest possible job start is too late"); 
			throw helper.getFault(); 
		}
		else {
			logger.debug("Reservation starts early enough"); 
		}
    }

  
    /**
     * Performs job submission. 
     * Validates job description(s). Uses the passed index servers URIs
     * (or default index servers), to find resources. 
     * Filters and selects resources based on job requirements and 
     * user preferences. 
     * @param jobDesc Job description(s) in DOM format. 
     * @param indexServerURIs URI to custom index servers to query. 
     * @param jobPreferences Preferences of job, including e.g. benchmarks
     * and resource selections algorithm. 
     * @return Job id for submitted job. 
     * @throws Exception
     */
    public SubmittedJob submitJob(JobDefinition_Type jobDescription, 
            String indexServerURIs[], 
            JobPreferencesType jobPreferences, 
            String clientURL, 
			boolean clientStagesInput) throws Exception {

        // Create instances of broker and ReservationUtil
        Broker broker = new Broker(this.numberOfThreads); 
        ReservationUtils reservationUtils = new ReservationUtils(); 
        broker.setReservationUtils(reservationUtils); 

        // Verify that job description contains all required 
        // attributes (currently, only testing for "Executable"). 
        // This call throws InvalidJobDescriptionFaultType upon error.
        broker.verify(jobDescription); 

        // Convert formats in job preferences, including
        // convertion of relative start delays to absolute time.
        JobPreferencesType prefs = null; 
        if(jobPreferences != null) {
            // prefs = fixPreferences(jobPreferences);        
        	prefs = SubmitterUtils.fixPreferences(jobPreferences); 
            if(prefs == null)
                logger.error("Error fixing job preferences");
        }
        else 
            logger.debug("No job preferences specified"); 
        
        // Perform resource discovery, and query resources of interest. 
        logger.info("Getting resources"); 
        InformationFinder informationFinder = informationFinderFactory.
            getInformationFinder(); 
        
     	// Check if submission target is defined in 
    	String resourceURIs[] = null; 
    	try {
    		resourceURIs = jobPreferences.getJobRequirement().getSubmissionTarget(); 
    	} catch (Exception e) {
    		resourceURIs = null; 
    	}    	         
        
        List foundResources = SubmitterUtils.getResources(informationFinder, 
				indexServerURIs, 
				resourceURIs, 
				broker); 
        
        // Filter list of retrieved resources and save those 
        // apt to execute the job. 
        List approvedResources = SubmitterUtils.filterResources(foundResources, 
        		jobDescription, 
				broker); 
     
        logger.info("Done getting resources."); 

        // Select resource for job and submit job. 
        List jobIds = new Vector(); 
        Object jobId = null;                

        logger.info("We have " + approvedResources.size() + " resources to select from"); 
        
        logger.info("Reordering resources"); 
        
        Selector selector = broker.getSelector(jobPreferences); 
        
        // Sort resources in decreasing order of aptness, 
        // this may include creating advance reservations.
        List orderedResources = broker.select(approvedResources,
        		jobDescription, 
				jobPreferences, 
				selector);
        logger.info("Done reordering resources"); 
        
        if(orderedResources.size() == 0) {
        	logger.error("No suitable resources found for job"); 
        	NoAptResourcesFoundFaultType f = new NoAptResourcesFoundFaultType(); 
        	FaultHelper helper = new FaultHelper(f); 
        	helper.setDescription("No suitable resources found for job"); 
        	throw helper.getFault(); 
        }        
        
        SubmittedJob result = new SubmittedJob(); 
        
        Dispatcher dispatchers[] = SubmitterUtils.getDispatchers(this.dispatcherClasses); 
        DataManager dataManagers[] = SubmitterUtils.getDataManagers(this.dataManagerClasses); 
        
        // Check is a latest job start was specified. 
        Calendar latestStart = null; 
        boolean useLatestStart = false; 
        if(jobPreferences != null) {
        	latestStart = jobPreferences.getJobRequirement().
				getLatestAbsoluteStart(); 
        	if(latestStart != null)
        		useLatestStart = true; 
        }
        
        // Try all suitable resources in decreased order of aptness
        GLUECERPType selectedResource = null;
        boolean done = false; 
        boolean failed = false; 
        ReservationInfo reservation = null; 
        while(!done) {
        	
        	try {
        		
        		logger.debug("There remains " + orderedResources.size() 
        				+ " resources to selected from"); 
        		selectedResource = (GLUECERPType) orderedResources.remove(0);
        		
        		// No more resources to select from. 
        		if(selectedResource == null) {
        			done = true; 
        			failed = true; 
        			continue; 
        		}
        		
        		logger.debug("Submitting job to: " 
        				+ selectedResource.getCluster(0).getSubCluster(0).
						getHost(0).getName()); 
        		
        		// Get reservation for job. 
        		reservation = reservationUtils.getReservationInfo(
        				(Object)selectedResource); 
        		
        		// ensure that latestStart is not too late...
        		if(useLatestStart) {
        			assertLatestStart(reservation, 
        					reservationUtils, 
							latestStart); 
        		}
        		
        		// Stage input files
        		if(!clientStagesInput) {
        			boolean staged = false; 
        			for(int i = 0; i < dataManagers.length; i++) {
        				logger.info("Staging input data using: " 
        						+ dataManagers[i].getClass()); 
        				try {
        					dataManagers[i].stageIn(selectedResource, 
        							jobDescription, 
									clientURL); 
        					staged = true; 
        				} catch (Exception e) {
        					logger.error("Could not stage input files using: " 
        							+ dataManagers[i].getClass()); 
        				}
        				logger.info("Done stating input data"); 
        			}
        			if(!staged) {
        				logger.error("Could not stage input data"); 
        				throw new Exception("Failed to stage input data"); 
        			}
        		} 
        		else {
        			logger.info("Ignoring input file staging (it is " 
        					+ "handled by the client)"); 
        		}
        		
        		boolean dispatched = false; 
        		for(int i = 0; i < dispatchers.length && !dispatched; i++) {
        			logger.info("Dispatching job using: " + 
        					dispatchers[i].getClass()); 
        			try {
        				jobId = dispatchers[i].requestJob(selectedResource, 
        						jobDescription, 
								reservation, // if null -> no reservation was created
								clientURL, 
								clientStagesInput); 
        				dispatched = true; 
        			} catch (Exception e) {        			
        				logger.error("Could not dispatch job using: " 
        						+ dispatchers[i].getClass()); 
        			}
        		}
        		if(!dispatched) {
        			logger.error("Could not dispatch job"); 
        			throw new Exception("Could not dispatch job"); 
        		}
        		
        		logger.info("Done dispatching job"); 
        		
        		if(reservation != null) {
        			logger.info("Confirming reservation"); 
        			try {
        				reservationUtils.confirmReservation(selectedResource); 
        			} catch (Exception e) {
        				logger.debug("Could not confirm reservation / "
        						+ " no reservation to confirm"); 
        			}
        			logger.info("Done confirming reservation"); 
        		}
        		
        		done = true; 
        		
        		// Remove those reservations that are not used. 
        		logger.debug("Releasing unused reservation"); 
        		reservationUtils.removeUnusedReservations(); 
        		
        		// Update informationfinder cache. 
        		informationFinder.getResourceInformationCache().update(
        				selectedResource, jobDescription); 					
        		
        	}
        	catch (Exception e) {
        		logger.error("Submission to " + selectedResource.
        				getCluster(0).getSubCluster(0).getHost(0).getName() 
						+ " failed: "); 
        		reservationUtils.removeReservation(selectedResource); 
        		if(orderedResources.size() == 0) {
        			logger.error("Failed to submit job"); 
        			failed = true; 
        			done = true; 
        			throw e; 
        		}
        	}
        }
        if(!failed) {
        	logger.info("Job submitted with jobid " + jobId); 
        	result.setLocalID(jobId.toString());
        	if(reservation != null) {
        		result.setReservationEPR(reservation.getEpr()); 
        	}
        	result.setJobDescription(jobDescription); 
        	return result; 
        }
        else {
        	JobRequestFaultType f = new JobRequestFaultType(); 
        	FaultHelper helper = new FaultHelper(f); 
        	helper.addDescription("All job submission attempts fail"); 
        	throw helper.getFault();
        }
    }

    /**
     *
     */
    public void shutdown() {
    }
    
    	/**
	 * @return Returns the dataManagerClasses.
	 */
	public String[] getDataManagerClasses() {
		return dataManagerClasses;
	}
	/**
	 * @return Returns the dispatcherClasses.
	 */
	public String[] getDispatcherClasses() {
		return dispatcherClasses;
	}
	/**
	 * @return Returns the informationFinderFactory.
	 */
	public InformationFinderFactory getInformationFinderFactory() {
		return informationFinderFactory;
	}
	/**
	 * @return Returns the numberOfThreads.
	 */
	public int getNumberOfThreads() {
		return numberOfThreads;
	}
    

}


